-module(wsHttpProtocol).

-include("eWSrv.hrl").
-include("wsCom.hrl").

-compile(inline).
-compile({inline_size, 128}).

-export([
   request/3
]).

-spec request(Stage :: stage(), Data :: binary(), State :: #state{}) -> {ok, NewState :: #state{}} | {done, NewState :: #state{}}  | {error, term()}.
request(reqLine, Data, State) ->
   case erlang:decode_packet(http_bin, Data, []) of
      {more, _} ->
         {ok, State#state{buffer = Data}};
      {ok, {http_request, Method, RawPath, Version}, Rest} ->
         case byte_size(Rest) > 0 of
            true ->
               request(header, Rest, State#state{stage = header, buffer = Rest, wsReq = #wsReq{method = Method, raw_path = RawPath, version = Version}, headerCnt = 0});
            _ ->
               {ok, State#state{stage = header, buffer = <<>>, wsReq = #wsReq{method = Method, raw_path = RawPath, version = Version}, headerCnt = 0}}
         end;
      {ok, {http_error, ErrStr}, _} ->
         {error, ErrStr};
      {ok, {http_response, _, _, _}, _} ->
         {error, http_response};
      {error, _Reason} = Ret ->
         Ret
   end;
request(header, Data, State) ->
   case erlang:decode_packet(httph_bin, Data, []) of
      {more, _} ->
         {ok, State#state{buffer = Data}};
      {ok, {http_header, _, Key, _, Value}, Rest} ->
         #state{headerCnt = HeaderCnt, temHeader = TemHeader, rn = Rn} = State,
         NewTemHeader = [{Key, Value} | TemHeader],
         NewHeaderCnt = HeaderCnt + 1,
         case NewHeaderCnt >= 100 of
            true ->
               {error, too_many_headers};
            _ ->
               case Key of
                  'Content-Length' ->
                     request(header, Rest, State#state{buffer = Rest, headerCnt = NewHeaderCnt, temHeader = NewTemHeader, contentLength = binary_to_integer(Value)});
                  'Transfer-Encoding' ->
                     case Value of
                        <<"chunked">> ->
                           case Rn of
                              undefined ->
                                 request(header, Rest, State#state{buffer = Rest, headerCnt = NewHeaderCnt, temHeader = NewTemHeader, contentLength = chunked, rn = binary:compile_pattern(<<"\r\n">>)});
                              _ ->
                                 request(header, Rest, State#state{buffer = Rest, headerCnt = NewHeaderCnt, temHeader = NewTemHeader, contentLength = chunked})
                           end;
                        _ ->
                           {error, 'Transfer-Encoding'}
                     end;
                  _ ->
                     request(header, Rest, State#state{buffer = Rest, headerCnt = NewHeaderCnt, temHeader = NewTemHeader})
               end
         end;
      {ok, http_eoh, Rest} ->
         #state{temHeader = TemHeader, contentLength = CLen, wsReq = WsReq} = State,
         NewWsReq = WsReq#wsReq{headers = TemHeader},
         case CLen of
            undefined ->
               {error, content_length};
            0 ->
               {done, State#state{buffer = Rest, wsReq = NewWsReq, temHeader = []}};
            _ ->
               case byte_size(Rest) > 0 of
                  true ->
                     request(body, Rest, State#state{stage = body, buffer = Rest, wsReq = NewWsReq, temHeader = []});
                  _ ->
                     {ok, State#state{stage = body, buffer = <<>>, wsReq = NewWsReq, temHeader = []}}
               end
         end;
      {ok, {http_error, ErrStr}, _Rest} ->
         {error, ErrStr};
      {error, _Reason} = Ret ->
         Ret
   end;
request(body, Data, State) ->
   #state{contentLength = CLen, wsReq = WsReq, temChunked = TemChunked, rn = Rn} = State,
   case CLen of
      chunked ->
         case parseChunks(Data, Rn, TemChunked) of
            {ok, NewTemChunked, Rest} ->
               {ok, State#state{buffer = Rest, temChunked = NewTemChunked}};
            {over, LastTemChunked, Rest} ->
               {done, State#state{buffer = Rest, temChunked = <<>>, wsReq = WsReq#wsReq{body = LastTemChunked}}};
            {error, _Reason} = Ret ->
               Ret
         end;
      _ ->
         BodySize = erlang:size(Data),
         if
            BodySize == CLen ->
               {done, State#state{buffer = <<>>, wsReq = WsReq#wsReq{body = Data}}};
            BodySize > CLen ->
               <<Body:CLen/binary, Rest/binary>> = Data,
               {done, State#state{buffer = Rest, wsReq = WsReq#wsReq{body = Body}}};
            true ->
               {ok, State#state{buffer = Data}}
         end
   end.

parseChunks(Data, Rn, Acc) ->
   case binary:split(Data, Rn) of
      [Size, Rest] ->
         case chunkSize(Size) of
            undefined ->
               {error, invalid_chunk_size};
            0 ->
               {over, Acc, Rest};
            HexSize ->
               case chunkBody(Rest, HexSize) of
                  not_enough_data ->
                     {ok, Acc, Data};
                  {ok, Body, Rest} ->
                     parseChunks(Rest, Rn, <<Acc/binary, Body/binary>>)
               end
         end
   end.

chunkBody(Data, Size) ->
   case Data of
      <<Body:Size/binary, "\r\n", Rest/binary>> ->
         {ok, Body, Rest};
      _ ->
         not_enough_data
   end.

chunkSize(Bin) ->
   try
      binary_to_integer(Bin, 16)
   catch
      error:badarg ->
         undefined
   end.